# coding=utf-8
"""
(C) FHCS GmbH

Generic ViledaUtils class implementing static methods to use for MS Excel import / export and API communication
"""

import logging
from datetime import datetime, timedelta
from os import walk
from traceback import format_exc
from typing import Any, Dict, List, Optional

from openpyxl import load_workbook

from .importsheetparser import ImportSheetParser
from .viledaexceptions import ImportCellParserException, ImportSheetParserException


class ViledaUtils:
    """
    Static class containig various methods to use for import / export, API communication, convertations
    """

    @staticmethod
    def loadexcels(xlsdir: str, tolog: logging.Logger, pfile: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        Loads all excels from given folder, parses all records in all sheets, and returns a list of dicts representing
        records found in sheets. From sheet to sheet records can contain different sets of fields, which is allowed.
        Parsing is done with ImportSheetParser class, with using field and row tags that have to be present in
        MS Excel sheets.

        :param xlsdir: path to directory with MS Excel *.xlsx workbooks
        :param tolog: logger object to use
        :param pfile: file to record import progress, one single file created, by default is "progress" in the
                      same xlsdir folder, or full path/filename can be specified explicitly by this parameter
        :return: list of task records loaded from excel sheets
        """
        # List of workbooks
        wbs = dict()
        # Name of progress record file - it must be ignored by the import
        if pfile is None:
            _pfile = "progress"
        else:
            _pfile = pfile
        # Load all files from folder
        _log_wbs_list = list()
        for _tmp_root, _tmp_dirs, _tmp_files in walk(xlsdir):
            for _tmp_name_walk_f in _tmp_files:
                try:
                    if _tmp_name_walk_f != _pfile:
                        wbs[_tmp_name_walk_f] = load_workbook(filename=f"{xlsdir.strip('/')}/{_tmp_name_walk_f}")
                        _log_wbs_list.append(_tmp_name_walk_f)
                except Exception:
                    tolog.warning(f"Can't load {xlsdir.strip('/')}/{_tmp_name_walk_f}:\n{format_exc()}")
        # Log loaded workbooks
        if _log_wbs_list:
            _log_str = (f"\nLoaded {len(_log_wbs_list)} "
                        f"MS Excel workbook{'s' if len(_log_wbs_list) != 1 else ''} from "
                        f"file{'s' if len(_log_wbs_list) > 1 else ''}: "
                        f"{', '.join([(chr(39) + _i + chr(39)) for _i in _log_wbs_list])}")
        else:
            _log_str = "\nNo MS Excel workbooks loaded"
        tolog.info(_log_str)
        # Full list for importing
        reclist = list()
        # Log record string
        _log_str_loading = "\nSummary:\n"
        # Run checks, append correct records to the import list
        for _wname, _w in wbs.items():
            _sheetnames = _w.sheetnames
            _loaded_count = 0
            _log_str_loading += f"   '{_wname}':\n"
            for _sname in _sheetnames:
                try:
                    _sheet_parser = ImportSheetParser(_w[_sname], _sname)
                    if not _sheet_parser.ok():
                        raise ImportSheetParserException(f"      ERROR: failed loading sheet '{_sname}'")
                    _parsed_records = _sheet_parser.getparsedrecords()
                    for _r in _parsed_records:
                        reclist.append({**_r, "_wname": _wname, "_sname": _sname})
                    _log_str_loading += _sheet_parser.getlog()
                    _log_str_loading += (f"      TOTAL:  parsed records {_sheet_parser.getpasrsestats()['OK']} "
                                         f"of {_sheet_parser.getpasrsestats()['TOTAL']}\n")
                    _loaded_count += 1
                except ImportSheetParserException as _e:
                    _log_str_loading += f"      {repr(_e)}\n"
                except ImportCellParserException as _e:
                    _log_str_loading += f"      {repr(_e)}\n"
            _log_str_loading += (f"      loaded {_loaded_count} sheet{'s' if _loaded_count != 1 else ''} "
                                 f"out of {len(_sheetnames)}\n")
            # And finally close the workbook
            _w.close()
        _log_str_loading += (f"   TOTAL {len(reclist)} "
                             f"record{'s' if len(reclist) != 1 else ''}\n")
        # Log records loading
        tolog.info(_log_str_loading)
        # Log full imported list
        _log_str_reclist = str()
        for _r in reclist:
            _log_str_reclist += str(_r) + "\n"
        tolog.debug(_log_str_reclist)
        # And finally return the list
        return reclist

    @staticmethod
    def jsondef(obj: Any) -> str:
        """
        :param obj: object unknown to json.dump
        :return: serialized value of object unknown to json.dump
        """
        if isinstance(obj, timedelta):
            return str(obj)
        elif isinstance(obj, datetime):
            return obj.strftime("%d.%m.%Y")
        return repr(obj)



# coding=utf-8
"""
(C) FHCS GmbH

Utilities classes and standalone functions to use for FA API interaction and communication
"""

from datetime import datetime
from json import dumps
from time import time
from typing import Any, Dict, List, Optional, Tuple, Union
from html import unescape

import aiohttp


def strnorm(strin: str) -> List[str]:
    """
    This does the following:

    -   removes from string all symbols except spaces, latin letters, including UTF-0080+, cyrillic letters
        and digits
    -   uppercases whole string
    -   splits by spaces into the list

    :param strin: given raw string
    :return: list representing the normalized string value
    """
    _ALLOWED = (" 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
                "¡¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿЀЁЂЃЄЅІЇЈЉЊЋЌЍЎЏ"
                "АБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩЪЫЬЭЮЯабвгдежзийклмнопрстуфхцчшщъыьэюя"
                "ѐёђѓєѕіїјљњћќѝўџѠѡѢѣѤѥѦѧѨѩѪѫѬѭѮѯѰѱѲѳѴѵѶѷѸѹѺѻѼѽѾѿҀҁҊҋҌҍҎҏҐґҒғҔҕҖҗҘҙҚқҜҝҞҟҠҡҢңҤҥҦҧҨҩҪҫҬҭ"
                "ҮүҰұҲҳҴҵҶҷҸҹҺһҼҽҾҿӀӁӂӃӄӅӆӇӈӉӊӋӌӍӎӏӐӑӒӓӔӕӖӗӘәӚӛӜӝӞӟӠӡӢӣӤӥӦӧӨөӪӫӬӭӮӯӰӱӲӳӴӵӶӷӸӹӺӻӼӽӾӿ")
    _orig = str(strin)
    _pre = str()
    for _i in _orig:
        if _i in _ALLOWED:
            _pre += _i
        else:
            _pre += chr(0x0020)
    return _pre.upper().split()


def getcookieval(httpsession: aiohttp.ClientSession, cname: str) -> Optional[str]:
    """
    Extract required cookie from aiohttp.CookieJar instance from aiohttp session.

    :param httpsession: Active aiohttp session to use
    :param cname: String name of cookie to extract
    :return: Value of cookie or None if no cookie existing
    """
    for c in httpsession.cookie_jar:
        if c.key == cname:
            return c.value
    return None


def gettag(htmltag: str) -> str:
    """
    Extract content of single HTML tag, normalize with strnorm()

    :param htmltag: String representing one single HTML tag
    :return: Value of tag content, if something wrong with parsing, return the given htmltag value
    """
    _c = htmltag
    _beg = htmltag.find(">")
    if not _beg < 0:
        _end = htmltag.find("<", _beg + 1)
        if not _end < 0:
            _c = htmltag[(_beg + 1):_end]
    if _c != htmltag:
        return chr(0x0020).join(strnorm(unescape(_c)))
    return unescape(htmltag)


async def fagetrefs(fadomain: str,
                    httpsession: aiohttp.ClientSession) -> Tuple[dict, dict, dict, dict, str]:
    """
    Gets users, sites, and spaces from FA API, normalizes all values with strnorm() and runs checks
    to compare API data with those records loaded from MS Excel sheets

    :param fadomain: FA domain name
    :param httpsession: Active aiohttp session to use
    :return: Tuple with four dicts containing normalized values for: user reference, site reference, floors and spaces
             references, and last tuple item is a string with debug log
    """
    # Constant, default SSL certificates check, False to disable certificates check
    _SSL = None
    # Debug log for output
    _debug_log = str()
    # Get all users
    r = await httpsession.post(f"https://{fadomain}.facilityapps.com/api/1.0/users_manager/data",
                               headers={"Accept":       "application/json",
                                        "X-CSRF-TOKEN": getcookieval(httpsession, "XSRF-TOKEN")
                                        },
                               params={"fields": FARequests.userquery()},
                               data=FARequests.userform(),
                               ssl=_SSL)
    _users_json = await r.json()
    _debug_log += f"Users received from API:\n{[_u['contact.name'] for _u in _users_json['data']]}" + "\n"
    # Extract our users from whole list
    _user_ref = dict()
    for _i in _users_json["data"]:
        # Normalise the user name with strnorm() and join resulting list with spaces
        _user_ref[chr(0x0020).join(strnorm(unescape(_i["contact.name"])))] = _i["id"]
    _debug_log += f"User reference built:\n{dumps(_user_ref, indent=3, ensure_ascii=False)}" + "\n"
    # Get all sites
    r = await httpsession.post(f"https://{fadomain}.facilityapps.com/api/1.0/site_manager/site_data",
                               headers={"Accept":       "application/json",
                                        "X-CSRF-TOKEN": getcookieval(httpsession, "XSRF-TOKEN")
                                        },
                               params={"fields": FARequests.sitequery()},
                               data=FARequests.siteform(),
                               ssl=_SSL)
    _sites_json = await r.json()
    _debug_log += f"Sites received from API:\n{[_s['name'] for _s in _sites_json['data']]}" + "\n"
    # Extract our sites from whole list
    _site_ref = dict()
    for _i in _sites_json["data"]:
        # Normalise the site name with strnorm() and join resulting list with spaces
        _site_ref[chr(0x0020).join(strnorm(unescape(_i["name"])))] = _i["id"]
    _debug_log += f"Sites reference built:\n{dumps(_site_ref, indent=3, ensure_ascii=False)}" + "\n"
    # Get all floors and spaces
    r = await httpsession.post(f"https://{fadomain}.facilityapps.com/api/1.0/floorplan/spaces/list",
                               headers={"Accept":       "application/json",
                                        "X-CSRF-TOKEN": getcookieval(httpsession, "XSRF-TOKEN")
                                        },
                               params={"fields": FARequests.spacesquery()},
                               data=FARequests.spacesform(),
                               ssl=_SSL)
    _spaces_json = await r.json()
    # Put IDs of floors and spaces in two reference dicts
    _floors_ref = dict()
    _spaces_ref = dict()
    # Logging counters
    _log_sites = 0
    _log_floors = 0
    _log_spaces = 0
    for _r in _spaces_json["data"]:
        if gettag(_r["site.name"]) not in _spaces_ref.keys():
            _floors_ref[gettag(_r["site.name"])] = dict()
            _spaces_ref[gettag(_r["site.name"])] = dict()
            _log_sites += 1
        if gettag(_r["floor.name"]) not in _spaces_ref[gettag(_r["site.name"])].keys():
            _floors_ref[gettag(_r["site.name"])][gettag(_r["floor.name"])] = unescape(_r["floor.id"])
            _spaces_ref[gettag(_r["site.name"])][gettag(_r["floor.name"])] = dict()
            _log_floors += 1
        if gettag(_r["name"]) not in _spaces_ref[gettag(_r["site.name"])][gettag(_r["floor.name"])].keys():
            _spaces_ref[gettag(_r["site.name"])][gettag(_r["floor.name"])][gettag(_r["name"])] = _r["id"]
            _log_spaces += 1
    _debug_log += f"Got from FA API sites: {_log_sites}, floors: {_log_floors}, spaces: {_log_spaces}" + "\n"
    _debug_log += f"Floors:\n{dumps(_floors_ref, indent=3, ensure_ascii=False)}" + "\n"
    _debug_log += f"Spaces:\n{dumps(_spaces_ref, indent=3, ensure_ascii=False)}" + "\n"
    # Return references
    return _user_ref, _site_ref, _floors_ref, _spaces_ref, _debug_log


class FARequests:
    """
    Reference with objects required for FA API requests
    """

    @staticmethod
    def getappversion() -> str:
        """
        :return: FA mobile app version
        """
        return "4.1.0"

    @staticmethod
    def userquery() -> str:
        """
        :return: URL query parameter value to use in request for user list through the API
        """
        return (r'["id","contact\\.name","contact\\.email","contact\\.phonenumber","roles\\.name",'
                r'"lastDevice\\.fapps_version","lastDevice\\.os_version","dummy_lastseen","disablelogin"]')

    @staticmethod
    def sitequery() -> str:
        """
        :return: URL query parameter value to use in request for site list through the API
        """
        return (r'["identifier","name","region\\.name","address\\.city","address\\.address",'
                r'"formSubmissionsToSend\\.id","logboo(kOpenTickets\\.id"]')

    @staticmethod
    def spacesquery() -> str:
        """
        :return: URL query parameter value to use in request for floors and spaces list through the API
        """
        return r'["nr","name","order","floor\\.name","site\\.name","floor\\.type"]'

    @staticmethod
    def tasksquery() -> str:
        """
        :return: URL query parameter value to use in request for full task list through the API
        """
        return (r'["title","current_status","owners","locations","elements",'
                r'"floor","space","date_start","date_end","date_end"]')

    @staticmethod
    def userform() -> dict:
        """
        :return: dict with application/x-www-form-urlencoded fields to use in request for user list through the API
        """
        return {"draw":                          r"5",
                "columns[0][data]":              r"select",
                "columns[0][name]":              r"",
                "columns[0][searchable]":        r"true",
                "columns[0][orderable]":         r"false",
                "columns[0][search][value]":     r"",
                "columns[0][search][regex]":     r"false",
                "columns[1][data]":              r"id",
                "columns[1][name]":              r"id",
                "columns[1][searchable]":        r"true",
                "columns[1][orderable]":         r"true",
                "columns[1][search][value]":     r"",
                "columns[1][search][regex]":     r"false",
                "columns[2][data]":              r"dummy_impersonate",
                "columns[2][name]":              r"dummy_impersonate",
                "columns[2][searchable]":        r"true",
                "columns[2][orderable]":         r"true",
                "columns[2][search][value]":     r"",
                "columns[2][search][regex]":     r"false",
                "columns[3][data]":              r"dummy_logout",
                "columns[3][name]":              r"dummy_logout",
                "columns[3][searchable]":        r"true",
                "columns[3][orderable]":         r"true",
                "columns[3][search][value]":     r"",
                "columns[3][search][regex]":     r"false",
                "columns[4][data]":              r"contact\.name",
                "columns[4][name]":              r"contact\.name",
                "columns[4][searchable]":        r"true",
                "columns[4][orderable]":         r"true",
                "columns[4][search][value]":     r"",
                "columns[4][search][regex]":     r"false",
                "columns[4][contact_name]":      r"true",
                "columns[5][data]":              r"contact\.email",
                "columns[5][name]":              r"contact\.email",
                "columns[5][searchable]":        r"true",
                "columns[5][orderable]":         r"true",
                "columns[5][search][value]":     r"",
                "columns[5][search][regex]":     r"false",
                "columns[6][data]":              r"contact\.phonenumber",
                "columns[6][name]":              r"contact\.phonenumber",
                "columns[6][searchable]":        r"true",
                "columns[6][orderable]":         r"true",
                "columns[6][search][value]":     r"",
                "columns[6][search][regex]":     r"false",
                "columns[7][data]":              r"roles\.name",
                "columns[7][name]":              r"roles\.name",
                "columns[7][searchable]":        r"true",
                "columns[7][orderable]":         r"true",
                "columns[7][search][value]":     r"",
                "columns[7][search][regex]":     r"false",
                "columns[7][concat][separator]": r"', '",
                "columns[8][data]":              r"lastDevice\.fapps_version",
                "columns[8][name]":              r"lastDevice\.fapps_version",
                "columns[8][searchable]":        r"true",
                "columns[8][orderable]":         r"true",
                "columns[8][search][value]":     r"",
                "columns[8][search][regex]":     r"false",
                "columns[9][data]":              r"lastDevice\.device_type",
                "columns[9][name]":              r"lastDevice\.device_type",
                "columns[9][searchable]":        r"true",
                "columns[9][orderable]":         r"true",
                "columns[9][search][value]":     r"",
                "columns[9][search][regex]":     r"false",
                "columns[10][data]":             r"lastDevice\.os_version",
                "columns[10][name]":             r"lastDevice\.os_version",
                "columns[10][searchable]":       r"true",
                "columns[10][orderable]":        r"true",
                "columns[10][search][value]":    r"",
                "columns[10][search][regex]":    r"false",
                "columns[11][data]":             r"dummy_lastseen",
                "columns[11][name]":             r"dummy_lastseen",
                "columns[11][searchable]":       r"true",
                "columns[11][orderable]":        r"true",
                "columns[11][search][value]":    r"",
                "columns[11][search][regex]":    r"false",
                "columns[12][data]":             r"disablelogin",
                "columns[12][name]":             r"disablelogin",
                "columns[12][searchable]":       r"true",
                "columns[12][orderable]":        r"true",
                "columns[12][search][value]":    r"",
                "columns[12][search][regex]":    r"false",
                "columns[13][data]":             r"displayColumns",
                "columns[13][name]":             r"",
                "columns[13][searchable]":       r"true",
                "columns[13][orderable]":        r"false",
                "columns[13][search][value]":    r"",
                "columns[13][search][regex]":    r"false",
                "order[0][column]":              r"1",
                "order[0][dir]":                 r"asc",
                "start":                         r"0",
                "length":                        str(2 ** 16),
                "search[value]":                 r"",
                "search[regex]":                 r"false"
                }

    @staticmethod
    def siteform() -> dict:
        """
        :return: dict with application/x-www-form-urlencoded fields to use in request for site list through the API
        """
        return {"draw":                      r"1",
                "columns[0][data]":          r"select",
                "columns[0][name]":          r"",
                "columns[0][searchable]":    r"true",
                "columns[0][orderable]":     r"false",
                "columns[0][search][value]": r"",
                "columns[0][search][regex]": r"false",
                "columns[1][data]":          r"id",
                "columns[1][name]":          r"id",
                "columns[1][searchable]":    r"true",
                "columns[1][orderable]":     r"true",
                "columns[1][search][value]": r"",
                "columns[1][search][regex]": r"false",
                "columns[2][data]":          r"identifier",
                "columns[2][name]":          r"identifier",
                "columns[2][searchable]":    r"true",
                "columns[2][orderable]":     r"true",
                "columns[2][search][value]": r"",
                "columns[2][search][regex]": r"false",
                "columns[3][data]":          r"name",
                "columns[3][name]":          r"name",
                "columns[3][searchable]":    r"true",
                "columns[3][orderable]":     r"true",
                "columns[3][search][value]": r"",
                "columns[3][search][regex]": r"false",
                "columns[4][data]":          r"region\.name",
                "columns[4][name]":          r"region\.name",
                "columns[4][searchable]":    r"true",
                "columns[4][orderable]":     r"true",
                "columns[4][search][value]": r"",
                "columns[4][search][regex]": r"false",
                "columns[5][data]":          r"address\.city",
                "columns[5][name]":          r"address\.city",
                "columns[5][searchable]":    r"true",
                "columns[5][orderable]":     r"true",
                "columns[5][search][value]": r"",
                "columns[5][search][regex]": r"false",
                "columns[6][data]":          r"address\.address",
                "columns[6][name]":          r"address\.address",
                "columns[6][searchable]":    r"true",
                "columns[6][orderable]":     r"true",
                "columns[6][search][value]": r"",
                "columns[6][search][regex]": r"false",
                "columns[7][data]":          r"formSubmissionsToSend\.id",
                "columns[7][name]":          r"formSubmissionsToSend\.id",
                "columns[7][searchable]":    r"true",
                "columns[7][orderable]":     r"true",
                "columns[7][search][value]": r"",
                "columns[7][search][regex]": r"false",
                "columns[7][allowed_join]":  r"true",
                "columns[8][data]":          r"logbookOpenTickets\.id",
                "columns[8][name]":          r"logbookOpenTickets\.id",
                "columns[8][searchable]":    r"true",
                "columns[8][orderable]":     r"true",
                "columns[8][search][value]": r"",
                "columns[8][search][regex]": r"false",
                "columns[9][data]":          r"displayColumns",
                "columns[9][name]":          r"",
                "columns[9][searchable]":    r"true",
                "columns[9][orderable]":     r"false",
                "columns[9][search][value]": r"",
                "columns[9][search][regex]": r"false",
                "order[0][column]":          r"1",
                "order[0][dir]":             r"asc",
                "start":                     r"0",
                "length":                    str(2 ** 16),
                "search[value]":             r"",
                "search[regex]":             r"false"
                }

    @staticmethod
    def spacesform() -> dict:
        """
        :return: dict with application/x-www-form-urlencoded fields to use in request for
                 floors and spaces through the API
        """
        return {"draw":                      r"1",
                "columns[0][data]":          r"select",
                "columns[0][name]":          r"",
                "columns[0][searchable]":    r"true",
                "columns[0][orderable]":     r"false",
                "columns[0][search][value]": r"",
                "columns[0][search][regex]": r"false",
                "columns[1][data]":          r"nr",
                "columns[1][name]":          r"nr",
                "columns[1][searchable]":    r"true",
                "columns[1][orderable]":     r"true",
                "columns[1][search][value]": r"",
                "columns[1][search][regex]": r"false",
                "columns[2][data]":          r"name",
                "columns[2][name]":          r"name",
                "columns[2][searchable]":    r"true",
                "columns[2][orderable]":     r"true",
                "columns[2][search][value]": r"",
                "columns[2][search][regex]": r"false",
                "columns[3][data]":          r"order",
                "columns[3][name]":          r"order",
                "columns[3][searchable]":    r"true",
                "columns[3][orderable]":     r"true",
                "columns[3][search][value]": r"",
                "columns[3][search][regex]": r"false",
                "columns[4][data]":          r"floor\.id",
                "columns[4][name]":          r"floor\.id",
                "columns[4][searchable]":    r"true",
                "columns[4][orderable]":     r"true",
                "columns[4][search][value]": r"",
                "columns[4][search][regex]": r"false",
                "columns[5][data]":          r"floor\.name",
                "columns[5][name]":          r"floor\.name",
                "columns[5][searchable]":    r"true",
                "columns[5][orderable]":     r"true",
                "columns[5][search][value]": r"",
                "columns[5][search][regex]": r"false",
                "columns[6][data]":          r"site\.id",
                "columns[6][name]":          r"site\.id",
                "columns[6][searchable]":    r"true",
                "columns[6][orderable]":     r"true",
                "columns[6][search][value]": r"",
                "columns[6][search][regex]": r"false",
                "columns[7][data]":          r"site\.name",
                "columns[7][name]":          r"site\.name",
                "columns[7][searchable]":    r"true",
                "columns[7][orderable]":     r"true",
                "columns[7][search][value]": r"",
                "columns[7][search][regex]": r"false",
                "columns[8][data]":          r"floor\.type",
                "columns[8][name]":          r"floor\.type",
                "columns[8][searchable]":    r"true",
                "columns[8][orderable]":     r"true",
                "columns[8][search][value]": r"",
                "columns[8][search][regex]": r"false",
                "order[0][column]":          r"1",
                "order[0][dir]":             r"asc",
                "start":                     r"0",
                "length":                    str(2 ** 16),
                "search[value]":             r"",
                "search[regex]":             r"false"
                }

    @staticmethod
    def tasksform(draw: int,
                  tbeg: datetime,
                  tend: datetime,
                  start: Optional[int] = None,
                  count: Optional[int] = None) -> dict:
        """
        Returns form fields to request full tasks list, for period from tbeg till tend,
        and pagination is also possible using start and count parameters, default is to return all available tasks.

        :param draw: draw ID, sequental
        :param tbeg: date and time from which to select tasks
        :param tend: date and time till which to select tasks
        :param start: if given, then the indexed starting task to return from
        :param count: if given, then it is the count of tasks to return
        :return: dict with application/x-www-form-urlencoded fields
        """
        if isinstance(start, int) and isinstance(count, int):
            _start = start
            _count = count
        else:
            _start = 0
            _count = 2 ** 31 - 1
        return {"draw":                       str(draw),
                "columns[0][data]":           r"select",
                "columns[0][name]":           r"",
                "columns[0][searchable]":     r"true",
                "columns[0][orderable]":      r"false",
                "columns[0][search][value]":  r"",
                "columns[0][search][regex]":  r"false",
                "columns[1][data]":           r"id",
                "columns[1][name]":           r"id",
                "columns[1][searchable]":     r"true",
                "columns[1][orderable]":      r"true",
                "columns[1][search][value]":  r"",
                "columns[1][search][regex]":  r"false",
                "columns[2][data]":           r"title",
                "columns[2][name]":           r"title",
                "columns[2][searchable]":     r"true",
                "columns[2][orderable]":      r"true",
                "columns[2][search][value]":  r"",
                "columns[2][search][regex]":  r"false",
                "columns[3][data]":           r"current_status",
                "columns[3][name]":           r"current_status",
                "columns[3][searchable]":     r"true",
                "columns[3][orderable]":      r"true",
                "columns[3][search][value]":  r"",
                "columns[3][search][regex]":  r"false",
                "columns[4][data]":           r"owners",
                "columns[4][name]":           r"owners",
                "columns[4][searchable]":     r"true",
                "columns[4][orderable]":      r"true",
                "columns[4][search][value]":  r"",
                "columns[4][search][regex]":  r"false",
                "columns[5][data]":           r"locations",
                "columns[5][name]":           r"locations",
                "columns[5][searchable]":     r"true",
                "columns[5][orderable]":      r"true",
                "columns[5][search][value]":  r"All",
                "columns[5][search][regex]":  r"false",
                "columns[6][data]":           r"elements",
                "columns[6][name]":           r"elements",
                "columns[6][searchable]":     r"true",
                "columns[6][orderable]":      r"true",
                "columns[6][search][value]":  r"",
                "columns[6][search][regex]":  r"false",
                "columns[7][data]":           r"floor",
                "columns[7][name]":           r"floor",
                "columns[7][searchable]":     r"true",
                "columns[7][orderable]":      r"true",
                "columns[7][search][value]":  r"",
                "columns[7][search][regex]":  r"false",
                "columns[8][data]":           r"space",
                "columns[8][name]":           r"space",
                "columns[8][searchable]":     r"true",
                "columns[8][orderable]":      r"true",
                "columns[8][search][value]":  r"",
                "columns[8][search][regex]":  r"false",
                "columns[9][data]":           r"hour_start",
                "columns[9][name]":           r"hour_start",
                "columns[9][searchable]":     r"true",
                "columns[9][orderable]":      r"true",
                "columns[9][search][value]":  r"",
                "columns[9][search][regex]":  r"false",
                "columns[10][data]":          r"minute_start",
                "columns[10][name]":          r"minute_start",
                "columns[10][searchable]":    r"true",
                "columns[10][orderable]":     r"true",
                "columns[10][search][value]": r"",
                "columns[10][search][regex]": r"false",
                "columns[11][data]":          r"hour_end",
                "columns[11][name]":          r"hour_end",
                "columns[11][searchable]":    r"true",
                "columns[11][orderable]":     r"true",
                "columns[11][search][value]": r"",
                "columns[11][search][regex]": r"false",
                "columns[12][data]":          r"minute_end",
                "columns[12][name]":          r"minute_end",
                "columns[12][searchable]":    r"true",
                "columns[12][orderable]":     r"true",
                "columns[12][search][value]": r"",
                "columns[12][search][regex]": r"false",
                "columns[13][data]":          r"date_start",
                "columns[13][name]":          r"date_start",
                "columns[13][searchable]":    r"true",
                "columns[13][orderable]":     r"true",
                "columns[13][search][value]": r"",
                "columns[13][search][regex]": r"false",
                "columns[14][data]":          r"date_end",
                "columns[14][name]":          r"date_end",
                "columns[14][searchable]":    r"true",
                "columns[14][orderable]":     r"true",
                "columns[14][search][value]": r"",
                "columns[14][search][regex]": r"false",
                "columns[15][data]":          r"date_end",
                "columns[15][name]":          r"date_end",
                "columns[15][searchable]":    r"true",
                "columns[15][orderable]":     r"true",
                "columns[15][search][value]": r"",
                "columns[15][search][regex]": r"false",
                "order[0][column]":           r"13",
                "order[0][dir]":              r"asc",
                "start":                      str(_start),
                "length":                     str(_count),
                "search[value]":              r"",
                "search[regex]":              r"false",
                "dates[from]":                tbeg.strftime("%Y-%m-%d"),
                "dates[to]":                  tend.strftime("%Y-%m-%d"),
                "dateColumn":                 r"date_end",
                "timeColumn":                 r"false"
                }

    @staticmethod
    def apploginform(usr: str, pwd: str) -> Dict[str, Any]:
        """
        :param usr: username string to use for login
        :param pwd: password string to use for login
        :return: dict with application/x-www-form-urlencoded fields to use for app API login request
        """
        return {"T": "LogIn_v2",
                "username": usr,
                "password": pwd,
                "session_id": str(),
                "version": FARequests.getappversion(),
                "mode": "LogIn",
                "auth_method": "basic"
                }

    @staticmethod
    def applogbookform(formdata: Union[Dict, List], siteid: int, sid: str) -> Dict[str, Any]:
        """
        :param formdata: JSON serializable object representing the form data
        :param siteid: integer site ID for FA API
        :param sid: session_id token got from login
        :return: dict with application/x-www-form-urlencoded fields to use for app API new logbook item submit
        """
        return {"T": "FormCheckList_v11",
                "L": "ru_RU",
                "FormData": dumps(formdata, ensure_ascii=False),
                "ReferenceId": str(siteid),
                "ReferenceType": "1",
                "Timestamp": str(int(time())),
                "session_id": sid,
                "AppVersionNumber": FARequests.getappversion(),
                "TaskData": "null"
                }

    @staticmethod
    def appchecklistsform(sid: str) -> Dict[str, Any]:
        """
        :param sid: session_id token got from login
        :return: dict with application/x-www-form-urlencoded fields to use for app API GetChecklists request
        """
        return {"session_id": sid,
                "T": "GetChecklists",
                "L": "ru_RU",
                "region_code": "ru_RU"
                }

    @staticmethod
    def appchecklistquestionsform(formids: List[int], sid: str) -> Dict[str, Any]:
        """
        :param formids: list of integer checklists IDs for FA API
        :param sid: session_id token got from login
        :return: dict with application/x-www-form-urlencoded fields to use for app API getChecklistQuestions request
        """
        return {"session_id": sid,
                "T": "getChecklistQuestions_v6",
                "checkListID": ";".join([str(_i) for _i in formids]),
                "image_data": "false",
                "region_code": "ru_RU"
                }



# coding=utf-8
"""
(C) FHCS GmbH

Custom exception classes
"""

class ImportCellParserException(Exception):
    """
    Raised when parsing specific errors in ImportCellParser occur
    """
    def __init__(self, message):
        self.message = message


class ImportSheetParserException(Exception):
    """
    Raised when parsing specific errors in ImportSheetParser occur
    """
    def __init__(self, message):
        self.message = message



# coding=utf-8
"""
viledatools
"""

__version__ = "1.0.19"

from typing import Tuple

from .apiutils import (FARequests as FARequests, fagetrefs as fagetrefs, getcookieval as getcookieval, gettag as gettag)
from .diagonaltablewalk import DiagonalTableWalk as DiagonalTableWalk
from .fatask import FATask as FATask
from .importcellparser import ImportCellParser as ImportCellParser
from .importsheetparser import ImportSheetParser as ImportSheetParser
from .viledaexceptions import (ImportCellParserException as ImportCellParserException,
                               ImportSheetParserException as ImportSheetParserException)
from .viledautils import ViledaUtils as ViledaUtils

__all__: Tuple[str, ...] = (
        "fagetrefs",
        "FARequests",
        "getcookieval",
        "gettag",
        "DiagonalTableWalk",
        "FATask",
        "ImportCellParser",
        "ImportSheetParser",
        "ImportCellParserException",
        "ImportSheetParserException",
        "ViledaUtils",
        )



# coding=utf-8
"""
(C) FHCS GmbH

Class parsing value from MS Excel cell given, depending on Excel field key provided
"""

import inspect
import types
from datetime import datetime, timedelta
from typing import Any, List, Optional, Tuple, Union

import openpyxl
from openpyxl.cell.cell import Cell

from .fatask import FATask
from .viledaexceptions import ImportCellParserException
from .apiutils import strnorm


# noinspection PyPep8Naming
class ImportCellParser:
    """
    Class parsing value from MS Excel cell given, depending on Excel field key provided
    """
    def __init__(self, fieldkey: Optional[str] = None):
        """
        Gets the Excel field key and provides parse method for that value

        :param fieldkey: case insensitive string field key as specified in the Excel sheet
        """
        # Get class parsers list for field keys, method name have to start with "IQF" same as field key in Excel
        self._parsers_list = [n.upper() for n, v in inspect.getmembers(self, inspect.ismethod)
                              if (isinstance(v, types.MethodType) and n.upper()[:3] == "IQF")]
        if fieldkey is not None:
            if fieldkey not in self._parsers_list:
                raise ImportCellParserException(f"Unknown field key: {fieldkey}")
            self._method = getattr(self, fieldkey)

    def parse(self, cellobject: openpyxl.cell.cell.Cell) -> Any:
        """
        Parse value given cell object with chosen key field

        :param cellobject: openpyxl cell object to parse
        :return: value parsed from the cell
        """
        return self._method(cellobject)

    def getfieldtags(self):
        """
        Returns list of field key tags
        """
        return self._parsers_list

    def _parsenames(self, cellobject: openpyxl.cell.cell.Cell, sep: str) -> Union[List[str], None]:
        """
        Parse list of task owners from cell value, use given separator string

        :param cellobject: openpyxl cell object to parse
        :param sep: string to use as separator for str.split()
        :return: normalized task owners' names in a list() parsed from the cell
        """
        _value = cellobject.value
        # Trim and split by semicolons
        _value_raw_list = _value.strip().split(sep)
        # Normalize each item and filter out empty ones, fill filtered list
        _value_users_list = list()
        for _item in _value_raw_list:
            _trimmed_item = _item.strip()
            if _trimmed_item:
                _value_users_list.append(chr(0x0020).join(strnorm(_trimmed_item)))
        return _value_users_list if _value_users_list else None

    def _parseyesno(self, cellobject: openpyxl.cell.cell.Cell) -> Union[bool, None]:
        """
        Parse "yes" / "no" type of values, it stringifies, trims, uppercases and matches with "yes" in
        the parsed list, in english, russian and german languages. So to return True this function have to find at least
        one "yes" in any of languages. If it is not found - it returns False.

        :param cellobject: openpyxl cell object to parse
        :return: True if cell value is parsed as yes, False otherwise
        """
        # Trim, uppercase and compare
        if str(cellobject.value).upper().strip() in ["ДА", "АГА", "Д", "YES", "YEP", "Y", "YEAH", "JA", "J"]:
            return True
        return False

    def _parseintseq(self, cellobject: openpyxl.cell.cell.Cell, allowedvals: List[int]) -> Union[List[str], None]:
        """
        Parse value of cell containing string with semicolon separated integers into a list of stringified integers,
        values in the MS Excel sheet must be given with ; (semicolon) separator

        :param cellobject: openpyxl cell object to parse
        :param allowedvals: list of allowed integers, if parsed integer not from this list, exception is raised
        :return: list of stringified integers parsed from the cell
        """
        # Trim and split by semicolons
        _value_raw_list = str(cellobject.value).strip().split(";")
        # Trim each item and filter out empty ones, fill filtered list
        _value_numbers_set = set()
        for _item in _value_raw_list:
            _trimmed_item = _item.strip()
            if _trimmed_item:
                try:
                    _trimmed_item_int = int(_trimmed_item)
                except Exception:
                    raise ImportCellParserException((f"_parseintseq(): Wrong value in the cell: "
                                                     f"{_trimmed_item}, values in the list must be integers divided by "
                                                     f"semicolons"))
                if _trimmed_item_int in allowedvals:
                    _value_numbers_set.add(str(_trimmed_item_int))
                else:
                    raise ImportCellParserException((f"_parseintseq(): Wrong value in the cell: "
                                                     f"{_trimmed_item_int}, integer values in the list must belong to "
                                                     f"{allowedvals}"))
        _value_numbers_list = list(_value_numbers_set)
        _value_numbers_list.sort()
        return _value_numbers_list

    def IQF001SITE(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Parse site name, normalize with strnorm()

        :param cellobject: openpyxl cell object to parse
        :return: normalized site name string parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return chr(0x0020).join(strnorm(cellobject.value))

    def IQF002FLOOR(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Parse floor name, normalize with strnorm()

        :param cellobject: openpyxl cell object to parse
        :return: normalized floor name parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return chr(0x0020).join(strnorm(cellobject.value))

    def IQF003SPACE(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Parse space name, normalize with strnorm()

        :param cellobject: openpyxl cell object to parse
        :return: normalized space name parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return chr(0x0020).join(strnorm(cellobject.value))

    def IQF004TASKOWNERS(self, cellobject: openpyxl.cell.cell.Cell) -> Union[List[str], None]:
        """
        Parse list of task owners from cell value, return list of normalized names with strnorm(). In the
        MS Excel sheet cell value user names must be separated with ; (semicolon) symbol, empty values between
        separators will be just ignored.

        :param cellobject: openpyxl cell object to parse
        :return: normalized task owners' names in a list() parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        if not isinstance(cellobject.value, str):
            raise ImportCellParserException((f"IQF004TASKOWNERS: Cell contents type is {type(cellobject.value)}, "
                                             f"string type required"))
        return self._parsenames(cellobject, ";")

    def IQF005TASKTIME(self, cellobject: openpyxl.cell.cell.Cell) -> Union[Tuple[timedelta, timedelta], None]:
        """
        Parse task start time and task end time into tuple

        :param cellobject: openpyxl cell object to parse
        :return: tuple (start time, end time) parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        _value = cellobject.value
        if not isinstance(_value, str):
            raise ImportCellParserException((f"IQF005TASKTIME: Cell contents type is {type(_value)}, "
                                             f"string type required"))
        _value_split = _value.strip().split("-")
        if len(_value_split) != 2:
            raise ImportCellParserException((f"IQF005TASKTIME: Wrong cell contents {_value.strip()}, "
                                             f"must be strictly in format H:M-H:M"))
        _output = []
        for _value_split_item in _value_split:
            _value_split_item_split = _value_split_item.strip().split(":")
            if len(_value_split_item_split) != 2:
                raise ImportCellParserException((f"IQF005TASKTIME: Wrong cell contents {_value.strip()}, "
                                                 f"must be strictly in format H:M-H:M"))
            _value_h = _value_split_item_split[0].strip()
            _value_m = _value_split_item_split[1].strip()
            if _value_h and _value_m:
                try:
                    _value_h_int = int(_value_h)
                    _value_m_int = int(_value_m)
                    if 0 <= _value_h_int <= 23 and 0 <= _value_m_int <= 59:
                        _output.append(timedelta(hours=_value_h_int, minutes=_value_m_int))
                    else:
                        raise Exception
                except Exception:
                    raise ImportCellParserException((f"IQF005TASKTIME: Wrong cell contents {_value.strip()}, "
                                                     f"must be strictly in format H:M-H:M"))
            else:
                raise ImportCellParserException((f"IQF005TASKTIME: Wrong cell contents {_value.strip()}, "
                                                 f"must be strictly in format H:M-H:M"))
        return _output[0], _output[1]

    def IQF006TASKTITLE(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Parse task title, apply spaces trimming

        :param cellobject: openpyxl cell object to parse
        :return: trimmed task title parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return str(cellobject.value).strip()

    def IQF007TASKDESCR(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Parse task description, apply spaces trimming

        :param cellobject: openpyxl cell object to parse
        :return: trimmed task description text parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return str(cellobject.value).strip()

    def IQF008STARTDATE(self, cellobject: openpyxl.cell.cell.Cell) -> Union[datetime, None]:
        """
        Parse starting date for task reccurence planning, and if the task is planned as standalone, this is the date
        when this task is to be planned

        :param cellobject: openpyxl cell object to parse
        :return: task start date parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        _value = cellobject.value
        if isinstance(_value, datetime):
            # Cell in MS Excel sheet formatted in date / time Excel format,
            # extract year, month, day from it as it is
            return datetime(_value.year, _value.month, _value.day)
        elif isinstance(_value, (int, float)):
            # Cell in MS Excel sheet formatted in numeric format, convert to the datetime
            if _value < 2:
                raise ImportCellParserException((f"IQF008STARTDATE: Start date provided as numeric value: {_value}, "
                                                 f"but the value is out of range, numerical date value must be not "
                                                 f"less than 2"))
            return datetime(1900, 1, 1) + timedelta(days=(_value - 2))
        elif isinstance(_value, str):
            # Cell in MS Excel sheet formatted as string, in this case it have to be formatted as %Y-%m-%d,
            # otherwise wrong result will be returned
            return datetime.strptime(_value, "%Y-%m-%d")

    def IQF009FREQTYPE(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Parse frequency type for task planning, apply spaces trimming and uppercase

        :param cellobject: openpyxl cell object to parse
        :return: trimmed & uppercased frequency type parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        _value_stripped_upper = str(cellobject.value).strip().upper()
        if _value_stripped_upper not in FATask.supportedfreqs():
            raise ImportCellParserException((f"IQF009FREQTYPE: Frequency type not supported by "
                                             f"parser: {_value_stripped_upper}"))
        return _value_stripped_upper

    def IQF010FREQINTVL(self, cellobject: openpyxl.cell.cell.Cell) -> Union[int, None]:
        """
        Parse frequency interval as integer, defines frequency of task planned recurrence

        :param cellobject: openpyxl cell object to parse
        :return: integer recurrence interval parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        _value = cellobject.value
        try:
            _value_int = int(_value)
        except Exception:
            raise ImportCellParserException(f"IQF010FREQINTVL: Wrong value in the cell: {_value}, must be integer")
        if _value_int < 1:
            raise ImportCellParserException((f"IQF010FREQINTVL: Value in the cell: {_value} out of range, "
                                             f"frequency interval must be greater than 0"))
        return _value_int

    def IQF011FREQDAYS(self, cellobject: openpyxl.cell.cell.Cell) -> Union[List[str], None]:
        """
        Parse task recurrence days as list of stringified integers, where ["1", "2", "3", "4", "5", "6", "7"]
        corresponds to [Mon, Tue, Wed, Thu, Fri, Sat, Sun], list values in the MS Excel sheet must
        be given with ; (semicolon) separator

        :param cellobject: openpyxl cell object to parse
        :return: list of stringified integers parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        if not isinstance(cellobject.value, (str, int)):
            raise ImportCellParserException((f"IQF011FREQDAYS: Cell contents type is {type(cellobject.value)}, "
                                             f"string type required"))
        return self._parseintseq(cellobject, [1, 2, 3, 4, 5, 6, 7])

    def IQF012ATMID(self, cellobject: openpyxl.cell.cell.Cell) -> Union[int, None]:
        """
        Made for mosklining domain

        Parse ATM ID, convert to int

        :param cellobject: openpyxl cell object to parse
        :return: int ATM ID parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        _value = cellobject.value
        try:
            _value_int = int(_value)
        except Exception:
            raise ImportCellParserException(f"IQF012ATMID: Wrong value in the cell: {_value}, must be integer")
        if _value_int < 0:
            raise ImportCellParserException((f"IQF012ATMID: Value in the cell: {_value} out of range, "
                                             f"frequency interval can not be less than 0"))
        return _value_int

    def IQF013ATMADDR(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Made for mosklining domain

        Parse address, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: stringified and trimmed cell value
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return str(cellobject.value).strip()

    def IQF014ATMISOFFICE(self, cellobject: openpyxl.cell.cell.Cell) -> Union[bool, None]:
        """
        Made for mosklining domain

        Parse yes or no if ATM is in the office, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: True if cell value is parsed as yes, False otherwise
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return self._parseyesno(cellobject)

    def IQF015ATMISPUBLIC(self, cellobject: openpyxl.cell.cell.Cell) -> Union[bool, None]:
        """
        Made for mosklining domain

        Parse yes or no if ATM is in publicly accessible area, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: True if cell value is parsed as yes, False otherwise
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return self._parseyesno(cellobject)

    def IQF016ATMISNFC(self, cellobject: openpyxl.cell.cell.Cell) -> Union[bool, None]:
        """
        Made for mosklining domain

        Parse yes or no if ATM has NFC functionality, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: True if cell value is parsed as yes, False otherwise
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return self._parseyesno(cellobject)

    def IQF017ATMMODEL(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Made for mosklining domain

        Parse ATM model string, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: stringified and trimmed cell value
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return str(cellobject.value).strip()

    def IQF018ATMCURR(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Made for mosklining domain

        Parse ATM available currencies description, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: stringified and trimmed cell value
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return str(cellobject.value).strip()

    def IQF019ATMSERIAL(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Made for mosklining domain

        Parse ATM serial number, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: stringified and trimmed cell value
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return str(cellobject.value).strip()

    def IQF020ATMCOMMENT(self, cellobject: openpyxl.cell.cell.Cell) -> Union[str, None]:
        """
        Made for mosklining domain

        Parse comment cell, trim spaces

        :param cellobject: openpyxl cell object to parse
        :return: stringified and trimmed cell value
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return str(cellobject.value).strip()

    def IQF021FREQMONTHS(self, cellobject: openpyxl.cell.cell.Cell) -> Union[List[str], None]:
        """
        Parse task recurrence months as list of stringified integers, where
        ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"] corresponds to
        [Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec], list values in the MS Excel sheet must
        be given with ; (semicolon) separator

        :param cellobject: openpyxl cell object to parse
        :return: list of stringified integers parsed from the cell
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        if not isinstance(cellobject.value, (str, int)):
            raise ImportCellParserException((f"IQF021FREQMONTHS: Cell contents type is {type(cellobject.value)}, "
                                             f"string type required"))
        return self._parseintseq(cellobject, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])

    def IQF022SAMEDAYOFWEEK(self, cellobject: openpyxl.cell.cell.Cell) -> Union[bool, None]:
        """
        If monthly recurrence is set, then this parameter defines whether task repeats same day of the week, or same
        date of each month

        If "yes" then task will repeat same day of the same week as first date of the task, if "no" - then task
        will repeat same date of the month

        :param cellobject: openpyxl cell object to parse
        :return: True if cell value is parsed as yes, False otherwise
        """
        if not isinstance(cellobject, Cell):
            raise TypeError(f"Provided cell object type {type(cellobject)}, type openpyxl.cell.cell.Cell required")
        if cellobject.value is None:
            return None
        return self._parseyesno(cellobject)



# coding=utf-8
"""
(C) FHCS GmbH

Class representing single FA task used for planning tasks through FA API
"""

from datetime import datetime, timedelta
from json import dumps
from typing import Dict, List, Optional, Union


class FATask:
    """
    Representing singe FA task for planning through FA API.
    """

    def __init__(self,
                 title: str,
                 details: str,
                 language: str,
                 site: int,
                 start: datetime,
                 end: datetime,
                 *,
                 owners: Optional[List[int]] = None,
                 approvers: Optional[List[int]] = None,
                 watchers: Optional[List[int]] = None,
                 spaces: Optional[Dict[str, List[int]]] = None,
                 duration: Optional[timedelta] = None) -> None:
        """
        Constructs the FA task object.

        :param title: Task title, UTF-8
        :param details: Task details string, may be multilene, UTF-8
        :param language: Task title and description language, "ru", "en", etc.
        :param site: Site ID in FA
        :param start: date and time when first occurence of task is planned to start, timezone information ignored
        :param end: date and time of deadline when first occurence is planned to copmlete, timezone information ignored
        :param owners: list of user IDs in FA to assign the task on, users will be responsible for tasks
        :param approvers: list of user IDs in FA who will become approvers, usually managers
        :param watchers: list of user IDs in FA who will have access to watch task status
        :param spaces: dict of spaces on which to plan the task, {"floor ID": [list of space IDs]}
        :param duration: duration of task as timedelta object
        """
        _d = (0, 0, 0)
        if duration is not None:
            (_dh, _ds) = divmod(duration.total_seconds(), 3600)
            _d = ((int(_dh),) + divmod(int(_ds), 60))
        self._task = {
                "advanced":                           "1",
                "id":                                 "",
                "task_title_id":                      "",
                "task_title":                         title,
                "description_field":                  details,
                "element":                            [],
                "date_start":                         start.strftime("%Y-%m-%d"),
                "hour_start":                         start.strftime("%H"),
                "minute_start":                       start.strftime("%M"),
                "date_end":                           end.strftime("%Y-%m-%d"),
                "hour_end":                           end.strftime("%H"),
                "minute_end":                         end.strftime("%M"),
                "locations":                          str(site),
                "owners":                             [str(_o) for _o in owners] if (owners is not None) else [],
                "duration_hour":                      str(_d[0]) if _d[0] != 0 else str(_d[0]).zfill(2),
                "duration_minute":                    str(_d[1]).zfill(2),
                "duration_second":                    str(_d[2]).zfill(2),
                "remove_upload":                      "",
                "repeat_interval_period":             "never",
                "repeat_interval_length":             "1",
                "frequency_monthly_type":             "monthly",
                "frequency_stop_repeat_number_value": "",
                "end_after_date":                     "",
                "frequency_stop_repeat":              "0",
                "sampling_select":                    "",
                "owner_roles":                        [],
                "approvers":                          [str(_a) for _a in approvers] if (approvers is not None) else [],
                "approver_roles":                     [],
                "watchers":                           [str(_w) for _w in watchers] if (watchers is not None) else [],
                "watcher_roles":                      [],
                "labels":                             [],
                "undefined":                          [],
                "task_sampling_select":               [],
                "task_complete_emails":               0,
                "task_canceled_emails":               0,
                "translations":                       {
                        f"{language.lower()}_{language.upper()}.utf8": title
                        },
                "description_translations":           {
                        f"{language.lower()}_{language.upper()}.utf8": details
                        },
                "floors_spaces":                      dict() if spaces is None else spaces
                }

    def __getitem__(self, item: str):
        """ Returns the task property by name """
        return self._task[item]

    def get(self, token: str) -> str:
        """
        Get the FA task as stringified JSON, given token value.

        :param token: FA API _token value
        :return: Full JSON as string, for POSTing to FA API
        """
        return dumps({"_token": token, **self._task}, ensure_ascii=False)

    def setrecurrence(self,
                      repeat: str,
                      *,
                      interval: Optional[int] = None,
                      dayset: Optional[List[str]] = None,
                      weekdayrepeat: Optional[bool] = None,
                      monthset: Optional[List[str]] = None,
                      end: Optional[Union[datetime, int]] = None,
                      ) -> None:
        """
        Sets the given recurrence pattern

        :param repeat: uppercased repetition frequency, "DAILY", "MONTHLY", and etc.
        :param interval: optional interval integer, 1 means repeat each period, 2 means one in two periods, and so on,
                         default is repeat each period
        :param dayset: optional list of strings representing day of the week to repeat, default is repeating each day,
                       for example ["1", "2", "7"], used only for "DAILY" frequency, ignored for other frequencies
        :param weekdayrepeat: optional only for monthly frequency, set to True to repeat monthly on the same day
                              of week as first task occurence, default is False which means repeat monthly the same day
                              of month; used only for "MONTHLY" frequency, ignored for other frequencies
        :param monthset: optional list of strings representing monthes to repeat, default is repeating each month,
                         for example ["3", "5", "6"]; used only for "MONTHLY" frequency, ignored for other frequencies
        :param end: optional datetime date after which stop task repetition, last repetition can happen on this date,
                    or if integer given then number of repetitions, default is None which means task is planned forever
        """
        # Check frequency
        if repeat in self.supportedfreqs():
            self._task["repeat_interval_period"] = repeat.lower()
        else:
            raise NotImplementedError(f"Not supported repetition frequency: {repeat}")
        # Check interval
        if interval is None:
            self._task["repeat_interval_length"] = str(1)
        else:
            if isinstance(interval, int) and interval > 0:
                self._task["repeat_interval_length"] = str(interval)
            else:
                raise RuntimeError(f"Wrong interval given: {interval}")
        # Check allowed days parameter
        if dayset is not None and repeat == "DAILY":
            if isinstance(dayset, list):
                for _i in dayset:
                    if not isinstance(_i, str) or int(_i) not in [1, 2, 3, 4, 5, 6, 7]:
                        raise RuntimeError(f"Wrong day given in day set: {_i}")
                _items_count = 0
                for _i in range(7):
                    if str(_i + 1) in dayset:
                        _items_count += 1
                if _items_count < 7:
                    self._task["frequency_daily_repeat"] = dayset
        # Check weekday repeat parameter
        if weekdayrepeat is not None and repeat == "MONTHLY":
            if isinstance(weekdayrepeat, bool):
                if weekdayrepeat:
                    self._task["frequency_monthly_type"] = "weekly"
            else:
                raise TypeError(f"Wrong value of weekdayrepeat given: {weekdayrepeat}")
        # Check allowed months parameter
        if monthset is not None and repeat == "MONTHLY":
            if isinstance(monthset, list):
                for _i in monthset:
                    if not isinstance(_i, str) or int(_i) not in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]:
                        raise RuntimeError(f"Wrong month given in month set: {_i}")
                _items_count = 0
                for _i in range(12):
                    if str(_i + 1) in monthset:
                        _items_count += 1
                if _items_count < 12:
                    self._task["frequency_monthly_repeat"] = monthset
        # Check end date or number repetitions
        if end is not None:
            if isinstance(end, datetime):
                self._task["end_after_date"] = end.strftime("%d-%m-%Y")
                self._task["frequency_stop_repeat"] = str(2)
            elif isinstance(end, int) and end > 0:
                self._task["frequency_stop_repeat_number_value"] = end
                self._task["frequency_stop_repeat"] = str(1)
            else:
                raise RuntimeError(f"Wrong end parameter given: {end}")

    @staticmethod
    def supportedfreqs() -> List[str]:
        """
        :return: list of supported recurrence frequencies
        """
        return ["DAILY", "MONTHLY"]



# coding=utf-8
"""
(C) FHCS GmbH

Iterator class, instantiated object returns (row, column) tuples to walk through tables starting from top left
corner down to the bottom right corner, to iterate through the table cells, for each next diagonal cell it returns
cells to the left of it ant to the top of it sequentally.

To use for MS Excel tables parsing, specifically for looking the main key tag of the import onm the sheet.
"""

from typing import Tuple


class DiagonalTableWalk:
    """
    Iterator class, instantiated object returns (row, column) tuples to walk through tables starting from top left
    corner down to the bottom right corner, to iterate through the table cells, for each next diagonal cell it returns
    cells to the left of it ant to the top of it sequentally. The algorithm does not search through entire table in
    case if table is not square - in this case the small tail of the table is skipped for sake of simplicity.
    Skipped tail size is not more than half of the remainder of division of bigger dimension by smaller dimension,
    like if table has 55 rows and 10 columns, then 25 (5x5) cells in the bottom right part of table will be skipped.
    """
    def __init__(self, maxrow: int, maxcol: int):
        """
        Pass limit values of the table to iterate through

        :param maxrow: the last table row to iterate through
        :param maxcol: the last table column to iterate through
        """
        self._maxrow = maxrow
        self._maxcol = maxcol
        # True if the table is non square and vertical, False if horizontal, None if table is square
        self._shape = None
        if self._maxrow > self._maxcol:
            self._shape = True
        elif self._maxrow < self._maxcol:
            self._shape = False
        # Shifts for non square tables - always start from top left, so 0 at the class instantiation
        self._rowshift = 0
        self._colshift = 0
        # Current row, start from top left
        self._row = None
        # Current column, start from top left
        self._col = None
        # Current diagonal bottom right corner, start from top left
        self._diag = 1
        # Current step from diagonal, start from 0, means first return the diagonal element itself,
        # it counts up to self._diag
        self._step = 0
        # Output trigger, True means return upper cursor, False means return left cursor, start with True
        self._cursor_switch = True

    def __next__(self) -> Tuple[int, int]:
        """
        Cacluates and returns the next tuple
        """
        # Updates the position of the cursor depending on the current self._diag,
        # self._step and self._cursor_switch values
        self._update()
        # And return the positioned cursor values
        return self._row, self._col

    def __iter__(self):
        """
        Returns self
        """
        return self

    def _update(self):
        """
        Positions the cursor depending on the switch
        """
        if self._step == 0:
            # If we are on the diagonal, position cursor on diagonal, this happens only on the first iteration
            self._row = self._rowshift + self._diag
            self._col = self._colshift + self._diag
            # and increment the step, because we have to increment distance of cursor from diagonal
            self._step += 1
        elif 0 < self._step < self._diag:
            # If we are not on the diagonal, but still inside the table
            if self._cursor_switch:
                # If self._cursor_switch is True, position cursor on the upper location,
                # at self._step distance from diagonal
                self._row = self._rowshift + self._diag - self._step
                self._col = self._colshift + self._diag
                # switch to False
                self._cursor_switch = False
            else:
                # If self._cursor_switch is False, position cursor on the left location,
                # at self._step distance from diagonal
                self._row = self._rowshift + self._diag
                self._col = self._colshift + self._diag - self._step
                # switch to True
                self._cursor_switch = True
                # and increment the step, because we have to increment distance of cursor from diagonal
                self._step += 1
        elif self._step == self._diag:
            # If we are not on the diagonal, and ran out of table,
            # increment the diagonal and set self._cursor_switch to True
            if self._check():
                self._diag += 1
                self._cursor_switch = True
                # And we are again on the diagonal, position cursor on diagonal
                self._row = self._rowshift + self._diag
                self._col = self._colshift + self._diag
                # Always drop the tail search
                if self._row > self._maxrow or self._col > self._maxcol:
                    raise StopIteration
                # drop the self._step to 1, for the next iteration
                self._step = 1
            else:
                raise StopIteration

    def _check(self) -> bool:
        """
        Checks max rows and columns to handle non square tables, and shifts the origin when needed
        """
        if self._shape is None:
            # Square table, just check limits for self._diag
            if self._diag < self._maxrow:
                return True
            else:
                return False
        else:
            # Table is not square
            if self._shape:
                # Table is vertical rows > columns
                if self._diag == self._maxcol:
                    # Shift row origin
                    self._rowshift += self._maxcol
                    # Reset self._diag to 0
                    self._diag = 0
            else:
                # Table is horizontal columns > rows
                if self._diag == self._maxrow:
                    # Shift column origin
                    self._colshift += self._maxrow
                    # Reset self._diag to 0
                    self._diag = 0
            # By default return True
            return True



# coding=utf-8
"""
viledatools

py setup.py sdist bdist_wheel
twine upload --repository pypi dist/*
"""

from setuptools import setup

setup(name='viledatools',
      version='1.0.19',
      packages=['viledatools'],
      url='https://www.vileda-professional.com/',
      license='Creative Commons Attribution 4.0 International',
      author='FHCS GmbH',
      author_email='support@fhcs.zendesk.com',
      description='Classes and functions to use for development of intregration software',
      install_requires=['aiohttp>=3.7.3', 'openpyxl>=3.0.6'],
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Other Environment',
                   'License :: Other/Proprietary License',
                   'Natural Language :: English',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.8',
                   'Topic :: Utilities'],
      zip_safe=False,
      python_requires='>=3.8'
      )



# coding=utf-8
"""
(C) FHCS GmbH

Class to parse values from MS Excel sheet given, this class defines and implements MS Excel importing rules
"""

from typing import Any, Dict, List, Tuple

import openpyxl
from openpyxl.utils import get_column_letter

from .diagonaltablewalk import DiagonalTableWalk
from .importcellparser import ImportCellParser
from .viledaexceptions import ImportCellParserException, ImportSheetParserException


class ImportSheetParser:
    """
    Class to parse values from MS Excel sheet given, this class defines and implements MS Excel importing rules
    """
    def __init__(self, ws: openpyxl.worksheet.worksheet.Worksheet, wsname: str):
        """
        Gets the imported worksheet object and provides methods for sheet parsing

        :param ws: openpyxl worksheet object to parse
        :param wsname: name of the sheet
        """
        # Local reference to the worksheet
        self._ws = ws
        # Parsed result list of dicts
        self._parsed: List[Dict[str, Any]] = list()
        # Parse log
        self._parse_log = str()
        # Parse OK flag:
        self._parse_ok = True
        # Parsed rows statistics
        self._rows_stats = {"TOTAL": 0, "OK": 0}
        # Row and column of main key tag IQXLSXTABLETAG
        self._mainkey: Tuple[int, int] = (0, 0)
        # Look for the main key
        if not self._find_mainkey():
            raise ImportSheetParserException(f"The main key IQXLSXTABLETAG not found in the excel sheet '{wsname}'")
        # Columns corresponding to field keys in imported MS Excel sheet
        self._columns: Dict[str, int] = dict()
        # Look for column field keys in the same row with main key tag
        self._find_columns()
        if not self._columns:
            # Not found field keys, raise
            raise ImportSheetParserException((f"Found no field key tags in main key row {self._mainkey[0]}, "
                                              f"nothing to import"))
        # Create parsers for each column
        self._p: Dict[str, ImportCellParser] = dict()
        self._register_parsers()
        # Parse
        self._parse_rows()

    def getparsedrecords(self) -> List[Dict[str, Any]]:
        """
        Returns the parsed result

        :return: List of dicts with parsed records for each line
        """
        return self._parsed

    def ok(self) -> bool:
        """
        Returns parse result

        :return: True if all parsed, False if any errors
        """
        return self._parse_ok

    def getlog(self) -> str:
        """
        Returns parse log

        :retrun: parse log
        """
        return self._parse_log

    def getpasrsestats(self) -> Dict[str, int]:
        """
        Returns parse statistics

        :retrun: parse statistics
        """
        return self._rows_stats

    def _find_mainkey(self) -> bool:
        """
        Looks for main key tag IQXLSXTABLETAG in the sheet
        """
        # Diagonal iterator to look for main key tag
        _twalk = DiagonalTableWalk(self._ws.max_row - self._ws.min_row + 1,
                                   self._ws.max_column - self._ws.min_column + 1)
        for _i in _twalk:
            if str(self._ws.cell(self._ws.min_row - 1 + _i[0],
                                 self._ws.min_column - 1 + _i[1]).value).strip() == "IQXLSXTABLETAG":
                # Redefine the self._mainkey with tuple (row, column) of the main key tag
                self._mainkey = (self._ws.min_row - 1 + _i[0], self._ws.min_column - 1 + _i[1])
                self._parse_log += (f"      OK: found main key "
                                    f"at {get_column_letter(self._mainkey[1])}{self._mainkey[0]}\n")
                return True
        return False

    def _find_columns(self):
        """
        Looks for columns containing field keys known to the parser
        """
        _fieldkeys_list = ImportCellParser().getfieldtags()
        for _col in range(self._ws.min_column, self._ws.max_column + 1):
            _col_value = str(self._ws.cell(self._mainkey[0], _col).value).strip()
            if _col_value in _fieldkeys_list:
                if _col_value not in self._columns.keys():
                    self._columns[_col_value] = _col
                    self._parse_log += (f"      OK: found known field key {_col_value} at "
                                        f"{get_column_letter(_col)}{self._mainkey[0]}\n")
                else:
                    self._parse_log += (f"      WARNING: ignored duplicated field key tag {_col_value} "
                                        f"in the cell {get_column_letter(_col)}{self._mainkey[0]}\n")

    def _register_parsers(self):
        """
        Creates parsers for each field key and for each column
        """
        for _field_key, _field_col in self._columns.items():
            try:
                self._p[_field_key] = ImportCellParser(_field_key)
                self._parse_log += f"      OK: registered parser for {_field_key}\n"
            except Exception as _e:
                self._parse_log += f"      ERROR: can not register parser for {_field_key}: {repr(_e)}\n"
                self._parse_ok = False

    def _parse_rows(self):
        """
        Parses all rows containing row tag IQXLSXROWTAG in the main key tag column
        """
        for _row in range(self._mainkey[0] + 1, self._ws.max_row + 1):
            # Get only rows with row key tag
            if str(self._ws.cell(_row, self._mainkey[1]).value) == "IQXLSXROWTAG":
                self._rows_stats["TOTAL"] += 1
                _parsed_row = dict()
                _parsed_ok = True
                for _field_key, _field_col in self._columns.items():
                    _cell_to_parse = self._ws.cell(_row, _field_col)
                    # Provide the openpyxl cell object to corresponding parser
                    try:
                        _parsed_row[_field_key] = self._p[_field_key].parse(_cell_to_parse)
                    except ImportCellParserException as _e:
                        self._parse_log += (f"      ERROR: row {_row} can't parse {_field_key} at "
                                            f"{get_column_letter(_field_col)}{_row}: {repr(_e)}\n")
                        _parsed_ok = False
                if _parsed_ok:
                    self._parsed.append(_parsed_row)
                    self._rows_stats["OK"] += 1
                else:
                    self._parse_log += f"      ERROR: row {_row} was not parsed\n"
